import docker
import click


class ContainerManager(object):
    def __init__(self,
                 docker_network_id=None,
                 skip_pull_image=False):

        self._docker_network_id = docker_network_id
        self._skip_pull_image = skip_pull_image
        self._docker_client = docker.from_env()

    def run(self, container):
        image = container.image

        if not self.has_image(image) or not self._skip_pull_image:
            self.pull_image(image)
        else:
            click.secho('skip pull image %s' % image)

        if not container.is_exist():
            container.create()

        container.start()

    def stop(self, container):
        pass

    def pull_image(self, image):
        try:
            click.secho('pull image %s...' % image, nl=False)
            for _ in self._docker_client.api.pull(image, stream=True, decode=True):
                click.secho('.', nl=False)
            click.secho('\n')
        except docker.errors.APIError as e:
            raise Exception('pull the docker image %s failed, %s' % (image, str(e)))

    def has_image(self, image):
        try:
            self._docker_client.images.get(image)
            return True
        except docker.errors.ImageNotFound:
            return False



import os
import logging
from tcfcli.libs.apis.provider import ApiProvider
from tcfcli.cmds.local.libs.apigw.local_service import LocalService, Route
from tcfcli.common.user_exceptions import NoApiDefinition

logger = logging.getLogger(__name__)


class LocalApiService(object):

    _DEFAULT_PORT = 3000
    _DEFAULT_HOST = '127.0.0.1'

    def __init__(self, invoke_context, port=None, host=None, static_dir=None):
        self._invoke_context = invoke_context
        self._stderr = invoke_context.stderr
        self._port = port or self._DEFAULT_PORT
        self._host = host or self._DEFAULT_HOST
        self._static_dir = static_dir

        self._local_runtime_manager = invoke_context.local_runtime_manager
        self._api_provider = ApiProvider(invoke_context.template)

    def start(self):
        routes_list = self._get_routes()

        if not routes_list:
            raise NoApiDefinition('There is no api definition in template')

        static_dir_path = self._get_static_dir_path()

        svc = LocalService(routes_list=routes_list,
                           runtime_manager=self._local_runtime_manager,
                           static_dir=static_dir_path,
                           port=self._port,
                           host=self._host,
                           stderr=self._stderr)

        self._show_routes(routes_list, port=self._port, host=self._host)
        logger.info('Mounting finsh. You can browse to the above endpoints to invoke functions. Support modify function code online, only need to restart tcf when the template changed.')

        svc.listen()

    def _get_routes(self):
        routes = []

        for api in self._api_provider.get_all():
            route = Route(method=[api.method], path=api.path, func_name=api.func_name)
            routes.append(route)

        return routes

    def _show_routes(self, routes_list, port, host):
        for route in routes_list:
            if route.method[0] is not 'ANY':
                logger.info('Mounting {} at http://{}:{}{} {}'.format(route.func_name, host, port, route.path, route.method))

    def _get_static_dir_path(self):
        if not self._static_dir:
            return None

        cwd = self._invoke_context.get_cwd()

        static_dir_path = os.path.join(cwd, self._static_dir)
        if os.path.exists(static_dir_path):
            return staticmethod

        return None



import click


class UserException(click.ClickException):
    exit_code = 1


class InvokeContextException(UserException):
    pass


class TemplateNotFoundException(UserException):
    pass


class OutTemplateAlreadyExist(UserException):
    pass


class ContextException(UserException):
    pass


class NoApiDefinition(UserException):
    pass

class InvalidTemplateException(UserException):
    pass

class UploadToCosFailed(UserException):
    pass


class FunctionNotFound(UserException):
    pass


class InvalidCodeDirs(UserException):
    pass


class UserConfigException(UserException):
    pass


class InvalidEnvParameters(UserException):
    pass


class InvalidOptionValue(UserException):
    pass


class InvalidResourceException(Exception):
    pass


class InvalidEventException(Exception):
    pass


class InvalidDocumentException(Exception):
    def __init__(self, causes):
        self._causes = sorted(causes)

    @property
    def message(self):
        return "Invalid Specification Document. Number of errors found: {}.".format(len(self.causes))

    @property
    def causes(self):
        return self._causes



from collections import namedtuple


Function = namedtuple("Function", [
    "name",
    "runtime",
    "memory",
    "timeout",
    "handler",
    "codeuri",
    "description",
    "cos_bucket_name",
    "cos_object_name",
    "zip_file",
    "environment",
    "vpc",
])

ResourcesKey = ("Handler",
                "Runtime",
                "Codeuri",
                "FunctionName",
                "Description",
                "MemorySize",
                "Timeout",
                "Environment",
                "ZipFile",
                "CodeBucket",
                "CodeObject",
                "VpcConfig")

GlobalsKey = ("Handler",
              "Runtime",
              "Description",
              "MemorySize",
              "Timeout",
              "Environment",
              "VpcConfig")

TENCENT_FUNCTION_TYPE = "TencentCloud::Serverless::Function"


class FunctionProvider(object):
    def get(self, name):
        raise NotImplementedError("Not Implemented")

    def get_all(self):
        raise NotImplementedError("Not Implemented")



from tcfcli.libs.tcsam.model import *
from tcfcli.libs.tcsam.events import *
from tcfcli.libs.tcsam.env import *
from tcfcli.libs.tcsam.vpc import *
from tcfcli.common.user_exceptions import InvalidTemplateException


class Function(ProperModel):
    PROPER = {
        "Handler": {
            MUST: True,
            TYPE: mstr,
            VALUE: None
        },
        "Runtime": {
            MUST: True,
            TYPE: mstr,
            VALUE: ("Python2.7", "Python3.6", "Nodejs6.10", "Nodejs8.9",
                    "Php5", "Php7", "Go1", "Java8")
        },
        "CodeUri": {
            MUST: False,
            TYPE: mstr,
            VALUE: None
        },
        "Description": {
            MUST: False,
            TYPE: mstr,
            VALUE: None
        },
        "MemorySize": {
            MUST: False,
            TYPE: mint,
            VALUE: None
        },
        "Timeout": {
            MUST: False,
            TYPE: mint,
            VALUE: False
        },
        "Environment": {
            MUST: False,
            TYPE: Environment,
            VALUE: None
        },
        "Events": {
            MUST: False,
            TYPE: Events,
            VALUE: None
        },
        "VpcConfig": {
            MUST: False,
            TYPE: VpcConfig,
            VALUE: None
        },
        "CosBucketName": {
            MUST: False,
            TYPE: mstr,
            VALUE: None
        },
        "CosObjectName": {
            MUST: False,
            TYPE: mstr,
            VALUE: None
        },
        "LocalZipFile": {
            MUST: False,
            TYPE: mstr,
            VALUE: None
        },
    }
    TYPE_PREFIX = "TencentCloud::Serverless::"

    def __init__(self, model, dft):
        if not isinstance(dft, dict):
            raise InvalidTemplateException("Globals invalid")
        if FUNCTION in dft:
            if not isinstance(dft[FUNCTION], dict):
                raise InvalidTemplateException("Globals invalid")
            for proper in dft[FUNCTION]:
                try:
                    setattr(self, proper, self.PROPER[proper][TYPE](dft[FUNCTION][proper]))
                except InvalidTemplateException as e:
                    raise InvalidTemplateException("{proper} ".format(proper=proper) + str(e))

        super(Function, self).__init__(model)


class Namespace(TypeModel):
    AVA_TYPE = {
        "TencentCloud::Serverless::Function": Function,
    }

    TYPE_PREFIX = "TencentCloud::Serverless::"

    def __init__(self, model, dft={}):
        super(Namespace, self).__init__(model, dft)


class Resources(TypeModel):
    AVA_TYPE = {
        "TencentCloud::Serverless::Namespace": Namespace,
    }

    def __init__(self, config):
        if not isinstance(config, dict):
            raise InvalidTemplateException("%s invalid" % self.__class__.__name__)
        resource = config.get(RESOURCE, {})
        if not resource:
            raise InvalidTemplateException("Resources not found")
        glo = config.get(GLOBAL, {})
        if not isinstance(resource, dict) or not isinstance(glo, dict):
            raise InvalidTemplateException("%s invalid" % self.__class__.__name__)

        super(Resources, self).__init__(resource, glo)

    def to_json(self):
        return {RESOURCE: self._serialize()}


if __name__ == '__main__':
    pass



from qcloud_cos import CosConfig
from qcloud_cos import CosS3Client
from tcfcli.common.user_config import UserConfig
from tcfcli.common.user_exceptions import UploadToCosFailed


class CosClient(object):
    def __init__(self):
        uc = UserConfig()
        self._config = CosConfig(Secret_id=uc.secret_id, Secret_key=uc.secret_key,
                                 Region=uc.region, Appid=uc.appid)
        self._client = CosS3Client(self._config)

    def upload_file2cos(self, bucket, file, key):
        # save funcs in the func directory
        try:
            response = self._client.put_object(Bucket=bucket,
                                               Body=file,
                                               Key=key,
                                               Metadata={
                                                   'x-cos-acl': 'public-read',
                                                   'Content-Type': 'application/x-zip-compressed',
                                                })
            if not response['ETag']:
                raise UploadToCosFailed("Upload func package failed")
        except Exception as e:
            raise e

        codeuri_in_cos = bucket + '/' + key
        return codeuri_in_cos



import click
from tcfcli.cmds.local.generate_event.generate_event_service import GenerateEventService


@click.command(name="generate-event", cls=GenerateEventService)
def generate_event():
    """
    \b
    Simulate a service generation event

    Common usage:

        \b
        Generate a cmq event:
        \b
        $ tcf local generate-event cmq notification  --owner  19911112
        \b
        Use the | to send to invoke
        \b
        tcf local generate-event cmq notification  --owner  19911112 | tcf local invoke -t template.yaml
    """
    pass



import click
import json
from tcfcli.cmds.cli import __version__
from tencentcloud.common import credential
from tencentcloud.common.profile.client_profile import ClientProfile
from tencentcloud.common.exception.tencent_cloud_sdk_exception import TencentCloudSDKException
from tencentcloud.scf.v20180416 import scf_client, models
from tcfcli.common.user_config import UserConfig
import base64


class ScfClient(object):

    def __init__(self):
        uc = UserConfig()
        self._cred = credential.Credential(secretId=uc.secret_id, secretKey=uc.secret_key)
        self._region = uc.region
        cp = ClientProfile("TC3-HMAC-SHA256")
        self._client = scf_client.ScfClient(self._cred, self._region, cp)
        self._client._sdkVersion = "TCFCLI"

    def get_function(self, function_name=None):
        req = models.GetFunctionRequest()
        req.FunctionName = function_name
        resp = self._client.GetFunction(req)
        return resp.to_json_string()

    def update_func_code(self, func, func_name, func_ns):
        req = models.UpdateFunctionCodeRequest()
        req.Namespace = func_ns
        req.FunctionName = func_name
        req.Handler = getattr(func, "Handler", None)
        req.ZipFile = self._model_zip_file(getattr(func, "LocalZipFile", None))
        req.CosBucketName = getattr(func, "CosBucketName", None)
        req.CosObjectName = getattr(func, "CosObjectName", None)
        resp = self._client.UpdateFunctionCode(req)
        return resp.to_json_string()

    def update_func_config(self, func, func_name, func_ns):
        req = models.UpdateFunctionConfigurationRequest()
        req.Namespace = func_ns
        req.FunctionName = func_name
        req.Description = getattr(func, "Description", None)
        req.MemorySize = getattr(func, "MemorySize", None)
        req.Timeout = getattr(func, "Timeout", None)
        req.Runtime = getattr(func, "Runtime", None)
        req.Environment = self._model_envs(getattr(func, "Environment", None))
        req.VpcConfig = self._model_vpc(getattr(func, "VpcConfig", None))
        resp = self._client.UpdateFunctionConfiguration(req)
        return resp.to_json_string()

    def create_func(self, func, func_name, func_ns):
        req = models.CreateFunctionRequest()
        req.Namespace =  func_ns
        req.FunctionName = func_name
        req.Handler = getattr(func, "Handler", None)
        req.Description = getattr(func, "Description", None)
        req.MemorySize = getattr(func, "MemorySize", None)
        req.Timeout = getattr(func, "Timeout", None)
        req.Runtime = getattr(func,"Runtime", None)
        req.Environment = self._model_envs(getattr(func, "Environment", None))
        req.VpcConfig = self._model_vpc(getattr(func, "VpcConfig", None))
        req.Code = self._model_code(getattr(func, "LocalZipFile", None),
                                    getattr(func, "CosBucketName", None),
                                    getattr(func, "CosObjectName", None))
        resp = self._client.CreateFunction(req)
        return resp.to_json_string()

    def deploy_func(self, func, func_name, func_ns, forced):
        try:
            self.create_func(func, func_name, func_ns)
            return
        except TencentCloudSDKException as err:
            if err.code in ["ResourceInUse.Function", "ResourceInUse.FunctionName"] and forced:
                pass
            else:
                return err
        click.secho("{ns} {name} already exists, update it now".format(ns=func_ns, name=func_name), fg="red")
        try:
            self.update_func_config(func, func_name, func_ns)
            self.update_func_code(func, func_name, func_ns)
        except TencentCloudSDKException as err:
            return err
        return

    def create_trigger(self, trigger, name, func_name, func_ns):
        req = models.CreateTriggerRequest()
        req.Namespace = func_ns
        req.FunctionName = func_name
        req.TriggerName = name
        req.Type = trigger.__class__.__name__.lower()
        trigger_desc = trigger.trigger_desc()
        if isinstance(trigger_desc, dict):
            trigger_desc = json.dumps(trigger_desc, separators=(',', ':'))
        req.TriggerDesc = trigger_desc
        req.Enable = getattr(trigger, "Enable", "OPEN")
        resp = self._client.CreateTrigger(req)
        click.secho(resp.to_json_string())

    def deploy_trigger(self, trigger, name, func_name, func_ns):
        try:
            self.create_trigger(trigger, name, func_name, func_ns)
        except TencentCloudSDKException as err:
            return err

    def deploy(self, func, func_name, func_ns, forced):
        err = self.deploy_func(func, func_name, func_ns, forced)
        if err:
            return err

    @staticmethod
    def _model_envs(environment):
        if environment is None:
            return None
        envs_array = []
        for k, v in vars(environment).items():
            var = models.Variable()
            var.Key = k
            var.Value = v
            envs_array.append(var)
        envi = models.Environment()
        envi.Variables = envs_array
        return envi

    @staticmethod
    def _model_vpc(vpc_config):
        if vpc_config:
            vpc = models.VpcConfig()
            vpc.VpcId = vpc_config.get("VpcId", None)
            vpc.SubnetId = vpc_config.get("SubnetId", None)
            return vpc
        return None

    @staticmethod
    def _model_code(zip_file, cos_buk_name, cos_obj_name):
        code = models.Code()
        code.CosBucketName = cos_buk_name
        code.CosObjectName = cos_obj_name
        if zip_file:
            with open(zip_file, 'rb') as f:
                code.ZipFile = base64.b64encode(f.read()).decode('utf-8')
        return code

    @staticmethod
    def _model_zip_file(zip_file):
        if zip_file:
            with open(zip_file, 'rb') as f:
                return base64.b64encode(f.read()).decode('utf-8')
        return None


import json
import sys
import os
import click
import subprocess
import threading
from tcfcli.libs.tcsam.tcsam import Resources
from tcfcli.libs.tcsam import model
from tcfcli.common.user_exceptions import InvokeContextException
from tcfcli.common.user_exceptions import InvalidOptionValue
from tcfcli.cmds.native.common.runtime import Runtime
from tcfcli.cmds.native.common.debug_context import DebugContext
from tcfcli.common.template import Template


class InvokeContext(object):

    BOOTSTRAP_SUFFIX = {
        "nodejs6.10": "bootstrap.js",
        "nodejs8.9": "bootstrap.js"
    }

    def __init__(self,
                 template_file,
                 function=None,
                 namespace=None,
                 debug_port=None,
                 debug_args="",
                 event="{}"):

        self._template_file = template_file
        self._function = function
        self._namespace = namespace
        self._event = event
        self._debug_port = debug_port
        self._debug_argv = debug_args
        self._runtime = None
        self._debug_context = None

    def _get_namespace(self, resource):
        ns = None
        if self._namespace:
            ns = resource.get(self._namespace, None)
        else:
            nss = list(resource.keys())
            if len(nss) == 1:
                self._namespace = nss[0]
                ns = resource.get(nss[0], None)
        if not ns:
            raise InvokeContextException("You must provide a valid namespace")

        del ns[model.TYPE]
        return ns

    def _get_function(self, namespace):
        fun = None
        if self._function:
            fun = namespace.get(self._function, None)
        else:
            funs = list(namespace.keys())
            if len(funs) == 1:
                self._function = funs[0]
                fun = namespace.get(funs[0], None)
        if not fun:
            raise InvokeContextException("You must provide a valid function")

        del fun[model.TYPE]
        return fun

    def __enter__(self):
        template_dict = Resources(Template.get_template_data(self._template_file)).to_json()
        resource = template_dict.get(model.RESOURCE, {})
        func = self._get_function(self._get_namespace(resource))
        self._runtime = Runtime(func[model.PROPERTY])
        self._debug_context = DebugContext(self._debug_port, self._debug_argv, self._runtime.runtime)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def invoke(self):
        def timeout_handle(child):
            try:
                click.secho('Function "%s" timeout after %d seconds' % (self._function, self._runtime.timeout))
                child.kill()
            except Exception:
                pass
        try:
            child = subprocess.Popen(args=[self.cmd]+self.argv, env=self.env)
        except OSError:
            click.secho("Execution failed,confirm whether the program({}) is installed".format(self._runtime.cmd))
        timer = threading.Timer(self._runtime.timeout, timeout_handle, [child])
        if not self._debug_context.is_debug:
            timer.start()
        child.wait()
        timer.cancel()

    @property
    def cmd(self):
        return self._debug_context.cmd if \
            self._debug_context.cmd is not None \
            else self._runtime.cmd

    @property
    def argv(self):
        argv = self._debug_context.argv
        runtime_pwd = os.path.dirname(os.path.abspath(__file__))
        bootstrap = os.path.join(runtime_pwd, "runtime", self._runtime.runtime,
                            self.BOOTSTRAP_SUFFIX[self._runtime.runtime])
        code = os.path.normpath(
            os.path.join(os.path.dirname(os.path.abspath(self._template_file)), self._runtime.codeuri))
        return argv + [bootstrap, os.path.join(code, self.get_handler())]

    @property
    def env(self):
        env = {
            'SCF_LOCAL': 'true',
            'SCF_FUNCTION_MEMORY_SIZE': str(self._runtime.mem_size),
            'SCF_FUNCTION_TIMEOUT': str(self._runtime.timeout),
            'SCF_EVENT_BODY': self._event,
            'SCF_FUNCTION_ENVIRON': json.dumps(self._runtime.env)
        }

        for k, v in self._runtime.env.items():
            env[k] = v

        for k, v in os.environ.items():
            env[k] = v

        return env

    def get_handler(self):
        res = self._runtime.handler.split('.', 1)
        if len(res) > 1:
            return res[0] + ':' + res[1]
        return res[0]



import click
from cookiecutter.main import cookiecutter
from cookiecutter import exceptions


class Init(object):

    RUNTIMES = {
        "python3.6": "gh:tencentyun/tcf-demo-python",
        "python2.7": "gh:tencentyun/tcf-demo-python",
        "go1": "gh:tencentyun/tcf-demo-go1",
        "php5": "gh:tencentyun/tcf-demo-php",
        "php7": "gh:tencentyun/tcf-demo-php",
        "nodejs6.10": "gh:tencentyun/tcf-demo-nodejs6.10",
        "nodejs8.9": "gh:tencentyun/tcf-demo-nodejs8.9",
        # "java8": "gh:tencentyun/tcf-demo-java8"
    }

    @staticmethod
    def do_cli(location, runtime, output_dir, name, no_input):

        click.secho("[+] Initializing project...", fg="green")
        params = {
            "template": location if location else Init.RUNTIMES[runtime],
            "output_dir": output_dir,
            "no_input": no_input,
        }
        click.secho("Template: %s" % params["template"])
        click.secho("Output-Dir: %s" % params["output_dir"])
        if not location and name is not None:
            params["no_input"] = True
            params['extra_context'] = {'project_name': name, 'runtime': runtime}
            click.secho("Project-Name: %s" % params['extra_context']["project_name"])
            click.secho("Runtime: %s" % params['extra_context']["runtime"])

        try:
            cookiecutter(**params)
        except exceptions.CookiecutterException as e:
            click.secho(e.message, fg="red")
            raise click.Abort()
        click.secho("[*] Project initialization is complete", fg="green")


@click.command()
@click.option('-l', '--location', help="Template location (git, mercurial, http(s), zip, path)")
@click.option('-r', '--runtime', type=click.Choice(Init.RUNTIMES.keys()), default="python3.6",
              help="Scf Runtime of your app")
@click.option('-o', '--output-dir', default='.', type=click.Path(), help="Where to output the initialized app into")
@click.option('-n', '--name',  default="demo", help="Name of your project to be generated as a folder")
@click.option('-N', '--no-input', is_flag=True, help="Disable prompting and accept default values defined template config")
def init(location, runtime, output_dir, name, no_input):
    """
    \b
    Initialize a Serverless Cloud Function with a scf template

    Common usage:

        \b
        Initializes a new scf using Python 3.6 default template runtime
        \b
        $ tcf init --runtime python3.6
        \b
        Initializes a new scf project using custom template in a Git repository
        \b
        $ tcf init --location gh:pass/demo-python
    """
    Init.do_cli(location, runtime, output_dir, name, no_input)



import click
from tcfcli.cmds.local.common.invoke_context import InvokeContext
from tcfcli.cmds.local.common.options import invoke_common_options
from tcfcli.common.user_exceptions import UserException


STD_IN = '-'


@click.command()
@click.option('--event', '-e', type=click.Path(), default=STD_IN)
@click.option('--no-event', is_flag=True, default=False)
@invoke_common_options
@click.argument('namespace_identifier', required=False, default='default')
@click.argument('function_identifier', required=False)
def invoke(template, namespace_identifier, function_identifier, event, no_event, env_vars, debug_port, debug_args, debugger_path,
           docker_volume_basedir, docker_network, log_file, skip_pull_image, region):
    '''
    \b
    Execute your scf in a docker environment locally
    \b
    Common usage:
        \b
        $ tcf local invoke -t template.yaml
    '''
    do_invoke(template, namespace_identifier, function_identifier, event, no_event, env_vars, debug_port, debug_args, debugger_path,
              docker_volume_basedir, docker_network, log_file, skip_pull_image, region)


def do_invoke(template, namespace_identifier, function_identifier, event, no_event, env_vars, debug_port, debug_args, debugger_path,
              docker_volume_basedir, docker_network, log_file, skip_pull_image, region):

    if no_event and event != STD_IN:
        raise UserException('event is conflict with no_event, provide only one.')

    if no_event:
        event_data = '{}'
    else:
        event_data = _get_event(event)

    try:
        with InvokeContext(template_file=template,
                           function_identifier=function_identifier,
                           env_vars_file=env_vars,
                           debug_port=debug_port,
                           debug_args=debug_args,
                           debugger_path=debugger_path,
                           docker_volume_basedir=docker_volume_basedir,
                           docker_network=docker_network,
                           log_file=log_file,
                           skip_pull_image=skip_pull_image,
                           region=region,
                           namespace=namespace_identifier) as context:

            context.local_runtime_manager.invoke(context.functions_name, event_data, context.stdout, context.stderr)

    except Exception as e:
        raise e


def _get_event(event_file):
    if event_file == STD_IN:
        click.secho('read event from stdin')

    with click.open_file(event_file, 'r') as f:
        return f.read()



EVENTS_METADATA = {
    "cmq": {
        "notification": {
            "filename": "notify",
            "help": "Generates an CMQ Topic notification event",
            "params": {
                "owner": {
                    "default": 12345678,
                }
            }
        }
    },
    "ckafka": {
        "consume": {
            "filename": "consume",
            "help": "Generates an ckafka consuming event",
            "params": {
            }
        }
    },
    "apigateway": {
        "proxy": {
            "filename": "proxy",
            "help": "Generates an API Gateway proxy event",
            "params": {
                "body": {
                    "default": "{\"msg\": \"hello apigateway\"}",
                }
            }
        }
    },
    "timer": {
        "timeup": {
            "filename": "timeout",
            "help": "Generates an Timer event",
            "params": {
                "message": {
                    "default": "hello timer",
                }
            }
        }
    },
    "cos": {
        "put": {
            "filename": "put",
            "help": "Generates an Cos Put event",
            "params": {
            }
        },
        "post": {
            "filename": "post",
            "help": "Generates an Cos Post event",
            "params": {
            }
        },
        "delete": {
            "filename": "delete",
            "help": "Generates an Cos Delete event",
            "params": {
            }
        }
    }
}



import click
from .invoke.cli import invoke
from tcfcli.cmds.local.generate_event.cli import generate_event


@click.group(name='native')
def native():
    """
    Run your scf natively for quick development
    """
    pass


native.add_command(invoke)
native.add_command(generate_event)





from tcfcli.libs.tcsam.model import *


class Timer(ProperModel):
    PROPER = {
        "CronExpression": {
            MUST: True,
            TYPE: mstr,
            VALUE: None
        },
        "Enable": {
            MUST: False,
            TYPE: mstr,
            VALUE: ["OPEN", "CLOSE"]
        }
    }

    def __init__(self, model, dft={}):
        super(Timer, self).__init__(model)

    def trigger_desc(self):
        return getattr(self, "CronExpression", None)


class COS(ProperModel):
    PROPER = {
        "Events": {
            MUST: True,
            TYPE: mstr,
            VALUE: None
        },
        "Filter": {
            MUST: False,
            TYPE: mmap,
            VALUE: None
        },
        "Enable": {
            MUST: False,
            TYPE: mstr,
            VALUE: ["OPEN", "CLOSE"]
        }
    }

    def __init__(self, model, dft={}):
        super(COS, self).__init__(model)

    def trigger_desc(self):
        return {"event": getattr(self, "Events"), "filter": getattr(self, "Filter", None)}


class APIGW(ProperModel):
    PROPER = {
        "StageName": {
            MUST: False,
            TYPE: mstr,
            VALUE: ["test", "prepub", "release"]
        },
        "HttpMethod": {
            MUST: False,
            TYPE: mstr,
            VALUE: ["ANY", "GET", "POST", "PUT", "DELETE", "HEAD"]
        }
    }

    def __init__(self, model, dft={}):
        super(APIGW, self).__init__(model)

    def trigger_desc(self):
        ret = {
            "api": {
                "authRequired": "FALSE",
                "requestConfig": {
                    "method": getattr(self, "HttpMethod", None)
                }
            },
            "service": {
                "serviceName": "SCF_API_SERVICE"
            },
            "release": {
                "environmentName": getattr(self, "StageName", None)
            }
        }
        return ret


class CMQ(ProperModel):
    PROPER = {
        "Enable": {
            MUST: False,
            TYPE: mstr,
            VALUE: ["OPEN", "CLOSE"]
        }
    }

    def __init__(self, model, dft={}):
        super(CMQ, self).__init__(model)

    def trigger_desc(self):
        return getattr(self, "Name", None)


class Events(TypeModel):
    AVA_TYPE = {
        "Timer": Timer,
        "COS": COS,
        "APIGW": APIGW,
        "CMQ": CMQ
    }

    def __init__(self, model, dft={}):
        super(Events, self).__init__(model, dft)

    def to_json(self):
        return self._serialize()


"""
Entry point for the CLI
"""

import click
import os
import sys
import logging

work_dir = os.getcwd()
dir_name, file_name = os.path.split(os.path.abspath(sys.argv[0]))
os.chdir(dir_name)
os.chdir('../../..')
sys.path.insert(0, os.getcwd())
os.chdir(work_dir)

from tcfcli.cmds.cli import __version__
from tcfcli.cmds.deploy.cli import deploy
from tcfcli.cmds.package.cli import package
from tcfcli.cmds.local.cli import local
from tcfcli.cmds.init.cli import init
from tcfcli.cmds.validate.cli import validate
from tcfcli.cmds.configure.cli import configure
from tcfcli.cmds.native.cli import native

logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')


@click.group()
@click.version_option(version=__version__, prog_name="TCF CLI")
def cli():
    pass


"""
Register commands into cli group
"""
cli.add_command(package)
cli.add_command(deploy)
cli.add_command(local)
cli.add_command(init)
cli.add_command(configure)
cli.add_command(validate)
cli.add_command(native)

if __name__ == "__main__":
    cli()



import os
import yaml
from tcfcli.common.user_exceptions import ContextException
from tcfcli.libs.function.fam_function_provider import ScfFunctionProvider
from tcfcli.libs.utils.yaml_parser import yaml_parse


class Context(object):
    def __init__(self, template_file=None, cos_bucket=None, output_template_file=None):
        self._template_file = template_file
        self._output_template_file = output_template_file
        self._cos_bucket = cos_bucket

        self._template_dict = None

    def __enter__(self):
        if self._template_file:
            self._template_dict = self.get_template_data(self._template_file)
            self._function_provider = ScfFunctionProvider(self._template_dict)
            self._functions = self._function_provider.get_all()
            self._deploy_template = self._function_provider.deploy_template

        return self

    def __exit__(self, *args):
        pass

    @property
    def template(self):
        return self._template_dict

    @property
    def template_path(self):
        return self._template_file

    @property
    def cos_bucket(self):
        return self._cos_bucket

    @property
    def output_template_file_path(self):
        return self._output_template_file

    @staticmethod
    def get_template_data(template_file):
        if not os.path.exists(template_file):
            return {}

        with open(template_file, 'r') as f:
            try:
                return yaml_parse(f.read())
            except (ValueError, yaml.YAMLError) as ex:
                raise ContextException("Parse template failed: {}".format(str(ex)))
        return

    def get_functions(self):
        for _, func in self._functions.items():
            yield func

    @property
    def deploy_template(self):
        return self._deploy_template



import click
from tcfcli.cmds.native.common.invoke_context import InvokeContext
from tcfcli.cmds.local.common.options import invoke_common_options
from tcfcli.common.user_exceptions import UserException

DEF_TMP_FILENAME = "template.yaml"

STD_IN = '-'


@click.command(name='invoke')
@click.option('--event', '-e', type=click.Path(), default=STD_IN)
@click.option('--no-event', is_flag=True, default=False)
@click.option('--template', '-t', default=DEF_TMP_FILENAME, type=click.Path(exists=True),
              envvar="TCF_TEMPLATE_FILE", show_default=True)
@click.option('--debug-port', '-d', help='The port exposed for debugging.', default=None)
@click.option('--debug-args', help='Additional args to be passed the debugger.', default="")
@click.argument('namespace_identifier', required=False, default='default')
@click.argument('function_identifier', required=False)
def invoke(template, namespace_identifier, function_identifier, event, no_event, debug_port, debug_args):
    '''
    \b
    Execute your scf in a environment natively
    \b
    Common usage:
        \b
        $ tcf native invoke -t template.yaml
    '''
    do_invoke(template, namespace_identifier, function_identifier, event, no_event, debug_port, debug_args)


def do_invoke(template, namespace, function, event, no_event, debug_port, debug_args):

    if no_event:
        event_data = "{}"
    else:
        click.secho('Enter a event:', color="green")
        with click.open_file(event, 'r') as f:
            event_data = f.read()

    try:
        with InvokeContext(
            template_file=template,
            namespace=namespace,
            function=function,
            debug_port=debug_port,
            debug_args=debug_args,
            event=event_data
        ) as context:
            context.invoke()
    except Exception as e:
        raise e


def _get_event(event_file):
    if event_file == STD_IN:
        click.secho('read event from stdin')

    with click.open_file(event_file, 'r') as f:
        return f.read()



import click
import sys
from tcfcli.libs.function.context import Context
from tcfcli.common.user_exceptions import TemplateNotFoundException, InvalidTemplateException
from tcfcli.libs.utils.scf_client import ScfClient
from tcfcli.libs.tcsam.tcsam import Resources


@click.command()
@click.option('--template-file', '-t', type=click.Path(exists=True), help="TCF template file for deploy")
@click.option('-f', '--forced', is_flag=True, default=False,
              help="Update the function when it already exists,default false")
def deploy(template_file, forced):
    '''
    Deploy a scf.
    '''
    deploy = Deploy(template_file, forced)
    deploy.do_deploy()


class Deploy(object):
    def __init__(self, template_file, forced=False):
        self.template_file = template_file
        self.check_params()
        self.resource = Resources(Context.get_template_data(self.template_file))
        self.forced = forced

    def do_deploy(self):
        for ns_name, ns in vars(self.resource).items():
            click.secho("deploy {ns} begin".format(ns=ns_name))
            for func_name, func in vars(ns).items():
                self._do_deploy_core(func, func_name, ns_name, self.forced)
            click.secho("deploy {ns} end".format(ns=ns_name))

    def check_params(self):
        if not self.template_file:
            click.secho("FAM Template Not Found", fg="red")
            raise TemplateNotFoundException("Missing option --template-file".format(self.template_file))

    def _do_deploy_core(self, func, func_name, func_ns, forced):
        err = ScfClient().deploy_func(func, func_name, func_ns, forced)
        if err is not None:
            if sys.version_info[0] == 3:
                s = err.get_message()
            else:
                s = err.get_message().encode("UTF-8")
            click.secho("Deploy function '{name}' failure. Error: {e}.".format(name=func_name,
                                                            e=s), fg="red")
            if err.get_request_id():
                click.secho("RequestId: {}".format(err.get_request_id().encode("UTF-8")), fg="red")
            return

        click.secho("Deploy function '{name}' success".format(name=func_name), fg="green")
        self._do_deploy_trigger(func, func_name, func_ns)

    def _do_deploy_trigger(self, func, func_name, func_ns):
        events = getattr(func, "Events", None)
        if events is None:
            return
        for name, trigger in vars(events).items():
            err = ScfClient().deploy_trigger(trigger, name, func_name, func_ns)
            if err is not None:
                if sys.version_info[0] == 3:
                    s = err.get_message()
                else:
                    s = err.get_message().encode("UTF-8")

                click.secho(
                    "Deploy trigger '{name}' failure. Error: {e}.".format(name=name,
                                                            e=s), fg="red")
                if err.get_request_id():
                    click.secho("RequestId: {}".format(err.get_request_id().encode("UTF-8")), fg="red")
                continue
            click.secho("Deploy trigger '{name}' success".format(name=name),fg="green")




"""
Wrapper to Docker Attach API
"""

import struct
from socket import timeout
from docker.utils.socket import read, read_exactly, SocketError


def attach(docker_client, container, stdout=True, stderr=True, logs=False):
    headers = {
        "Connection": "Upgrade",
        "Upgrade": "tcp"
    }

    query_params = {
        "stdout": 1 if stdout else 0,
        "stderr": 1 if stderr else 0,
        "logs": 1 if logs else 0,
        "stream": 1,
        "stdin": 0,
    }

    api_client = docker_client.api

    url = "{}/containers/{}/attach".format(api_client.base_url, container.id)

    response = api_client._post(url, headers=headers, params=query_params, stream=True)
    socket = api_client._get_raw_response_socket(response)

    return _read_socket(socket)


def _read_socket(socket):
    while True:

        try:

            payload_type, payload_size = _read_header(socket)
            if payload_size < 0:
                break

            for data in _read_payload(socket, payload_size):
                yield payload_type, data

        except timeout:
            pass
        except SocketError:
            break


def _read_payload(socket, payload_size):
    remaining = payload_size
    while remaining > 0:

        data = read(socket, remaining)
        if data is None:
            continue

        if len(data) == 0:
            break

        remaining -= len(data)
        yield data


def _read_header(socket):
    data = read_exactly(socket, 8)

    return struct.unpack('>BxxxL', data)



from tcfcli.libs.tcsam.model import *
from tcfcli.common.user_exceptions import InvalidTemplateException


class Runtime(object):
    RUNTIME = {
        'nodejs6.10': "node",
        'nodejs8.9': "node"
    }

    def __init__(self, proper):
        self.codeuri = proper.get(CODE_URI)
        self.env = proper.get(ENVIRONMENT, {VARIABLE: {}}).get(VARIABLE, {})
        self.handler = proper.get(HANDLER, "")
        self.mem_size = proper.get(CODE_URI, 128)
        self.runtime = proper.get(RUNTIME, "").lower()
        self.timeout = proper.get(TIMEOUT, 3)

        if self.runtime not in self.RUNTIME.keys():
            raise InvalidTemplateException("Invalid runtime. supports [{}]".
                                           format(",".join(self.RUNTIME.keys())))

    @property
    def cmd(self):
        return self.RUNTIME[self.runtime]



class FunctionConfig(object):
    _DEFAULT_TIMOUT_SECONDS = 3
    _DEFAULT_MEMORY = 128

    def __init__(self,
                 name,
                 runtime,
                 handler,
                 code_uri,
                 memory=None,
                 timeout=None,
                 envs=None,
                 vpc=None):
        self.name = name
        self.runtime = runtime
        self.handler = handler
        self.code_uri = code_uri
        self.memory = memory or self._DEFAULT_MEMORY
        self.timeout = timeout or self._DEFAULT_TIMOUT_SECONDS

        if not envs:
            # future set default
            pass

        if not vpc:
            pass


class EnviromentVariables(object):
    pass


class VpcConfigs(object):
    pass



from tcfcli.libs.tcsam.model import *


class VpcConfig(ProperModel):
    PROPER = {
        "VpcId": {
            MUST: True,
            TYPE: mstr,
            VALUE: None
        },
        "SubnetId": {
            MUST: True,
            TYPE: mstr,
            VALUE: None
        },

    }

    TYPE_PREFIX = "TencentCloud::Serverless::"

    def __init__(self, model):
        super(VpcConfig, self).__init__(model)

    def to_json(self):
        return self._serialize()



from flask import make_response, jsonify


class ErrorResponse(object):

    _INTERNAL_ERROR = {'errorMessage': 'Internal error'}
    _NOT_FOUND_BIND_FUNC = {'errorMessage': 'Not found the function bind to the route'}
    _INVALID_RESPONSE_FORMAT = {'errno': 403, 'error': 'Invalid scf response format. please check your scf response format.'}

    _STATUS_CODE_403 = 403
    _STATUS_CODE_502 = 502

    @staticmethod
    def InternalError(msg=None):
        err_msg = ErrorResponse._INTERNAL_ERROR

        if msg and isinstance(msg, str):
            err_msg = {'errorMessage': msg}

        return make_response(jsonify(err_msg), ErrorResponse._STATUS_CODE_502)

    @staticmethod
    def FuncNotFound(msg=None):
        err_msg = ErrorResponse._NOT_FOUND_BIND_FUNC

        if msg and isinstance(msg, str):
            err_msg = {'errorMessage': msg}

        return make_response(jsonify(err_msg), ErrorResponse._STATUS_CODE_403)

    @staticmethod
    def InvalidResponseFormat(msg=None):
        err_msg = ErrorResponse._INVALID_RESPONSE_FORMAT

        if msg and isinstance(msg, str):
            err_msg = {'errorMessage': msg}

        return make_response(jsonify(err_msg), ErrorResponse._STATUS_CODE_403)


import os
import posixpath
import re
try:
    import pathlib
except ImportError:
    import pathlib2 as pathlib


def to_posix_path(code_path):
    """
    Change the code_path to be of unix-style if running on windows when supplied with an absolute windows path.

    Parameters
    ----------
    code_path : str
        Directory in the host operating system that should be mounted within the container.
    Returns
    -------
    str
        Posix equivalent of absolute windows style path.
    Examples
    --------
    >>> to_posix_path('/Users/UserName/app')
    >>> to_posix_path('C:\\\\Users\\\\UserName\\\\AppData\\\\Local\\\\Temp\\\\mydir')
    """

    return re.sub("^([A-Za-z])+:",
                  lambda match: posixpath.sep + match.group().replace(":", "").lower(),
                  pathlib.PureWindowsPath(code_path).as_posix()) if os.name == "nt" else code_path



import os


class DebugContext(object):

    def __init__(self,
                 debug_port=None,
                 debugger_path=None,
                 debug_args=None):

        self.debug_port = debug_port
        self.debugger_path = debugger_path
        self.debug_args = debug_args
        if self.debug_port:
            os.environ["PYTHONUNBUFFERED"] = "1"

    def __bool__(self):
        return bool(self.debug_port)

    def __nonzero__(self):
        return self.__bool__()



import click
import base64
import os
import json
from tcfcli.libs.utils import py_version
from chevron import renderer


class GenerateEventAction(click.MultiCommand):
    '''
    this is a test
    '''

    PARAMS = "params"

    def __init__(self, service, srv_info, *args, **kwargs):
        super(GenerateEventAction, self).__init__(*args, **kwargs)
        self.service = service
        self.srv_info = srv_info

    def get_command(self, ctx, cmd):

        if cmd not in self.srv_info:
            return None

        params = []
        for param in self.srv_info[cmd][self.PARAMS].keys():
            default = self.srv_info[cmd][self.PARAMS][param]["default"]
            params.append(click.Option(
                ["--{}".format(param)],
                default=default,
                help="Specify the {}, default is {}".format(param, default)
            ))

        cbfun = click.pass_context(self.action)
        cmd = click.Command(name=cmd,
                            short_help=self.srv_info[cmd]["help"],
                            params=params, callback=cbfun)

        return cmd

    def list_commands(self, ctx):
        return sorted(self.srv_info)

    def action(self, ctx, *args, **params):
        action = ctx.info_name
        params_tmp = self.srv_info[action][self.PARAMS]

        params = self._param_encode(params,  params_tmp)

        pwd = os.path.dirname(os.path.abspath(__file__))
        filename = self.srv_info[action]["filename"] + ".js"
        path = os.path.join(pwd, "events", self.service, filename)

        with open(path) as f:
            data = f.read().strip()

        data = renderer.render(data, params)
        click.secho(json.dumps(json.loads(data), indent=2))

    def _param_encode(self, params, param_tmp):
        for p in params:
            encoding = param_tmp[p].get("encoding")
            if encoding == "url":
                params[p] = py_version.url_encoding(params[p])
            elif encoding == "base64":
                params[p] = base64.b64encode(params[p].encode('utf8')).decode('utf-8')
        return params



import logging
import copy
from tcfcli.libs.function.provider import FunctionProvider, TENCENT_FUNCTION_TYPE, Function
from tcfcli.libs.function.wrapper import TemplateReimplemented

logger = logging.getLogger(__name__)


class ScfFunctionProvider(FunctionProvider):

    _TENCENT_FUNCTION = TENCENT_FUNCTION_TYPE
    _DEFAULT_CODEURI = "."
    _DEFAULT_RUNTIME = "python2.7"
    _DEFAULT_HANDLER = "index.main_handler"
    _DEFAULT_TIMOUT_SECONDS = 3
    _DEFAULT_MEMORY = 128

    def __init__(self, template_dict):
        self.template_dict = self.get_final_template(template_dict)
        self.resources = self.template_dict.get("Resources", {})
        if len(self.resources) > 1:
            logger.debug("There are %d function resources found in the template", len(self.resources))

        self.functions = self._extract_functions(self.resources)

    def get(self, name):
        if not name:
            raise ValueError("Function name is required")

        return self.functions.get(name)

    def get_all(self):
        return self.functions

    def get_functions(self):
        for _, func in self.functions.items():
            yield func

    @staticmethod
    def get_final_template(template_dict):
        template_dict = template_dict or {}
        # if template_dict:
        #    template_dict = TemplateReimplemented(template_dict).parse()
        return template_dict

    @staticmethod
    def _extract_functions(resources):
        result = {}

        for name, resource in resources.items():
            resource_properties = resources[name].get("Properties", {})
            result[name] = ScfFunctionProvider._convert_fam_function_resource(name, resource_properties)

        return result

    @staticmethod
    def _convert_fam_function_resource(name, resource_properties):
        codeuri = resource_properties.get("CodeUri", ScfFunctionProvider._DEFAULT_CODEURI)
        cos_bucket_name = resource_properties.get("CosBucketName")
        cos_object_name = resource_properties.get("CosObjectName")
        zip_file = resource_properties.get("LocalZipFile")

        return Function(
            name=name,
            runtime=resource_properties.get("Runtime", ScfFunctionProvider._DEFAULT_RUNTIME),
            memory=resource_properties.get("MemorySize", ScfFunctionProvider._DEFAULT_MEMORY),
            timeout=resource_properties.get("Timeout", ScfFunctionProvider._DEFAULT_TIMOUT_SECONDS),
            handler=resource_properties.get("Handler", ScfFunctionProvider._DEFAULT_HANDLER),
            description=resource_properties.get("Description"),
            codeuri=codeuri,
            cos_bucket_name=cos_bucket_name,
            cos_object_name=cos_object_name,
            zip_file=zip_file,
            environment=resource_properties.get("Environment"),
            vpc=resource_properties.get("VpcConfig")
        )

    @property
    def deploy_template(self):
        template_dict = copy.deepcopy(self.template_dict)
        if "Globals" in template_dict:
            del template_dict["Globals"]

        return template_dict



import click
from .invoke.cli import invoke
from .start_api.cli import start_api
from .generate_event.cli import generate_event

@click.group(name='local')
def local():
    """
    Run your scf locally for quick development
    """
    pass


local.add_command(invoke)
# local.add_command(start_api)
local.add_command(generate_event)




import os
import copy
import click
import uuid
from zipfile import ZipFile, ZIP_DEFLATED
from io import BytesIO
from tcfcli.libs.utils.yaml_parser import yaml_dump
from tcfcli.libs.function.context import Context
from tcfcli.libs.utils.cos_client import CosClient
from tcfcli.common.user_exceptions import TemplateNotFoundException, \
    InvalidTemplateException, ContextException
from tcfcli.libs.tcsam.tcsam import Resources

_DEFAULT_OUT_TEMPLATE_FILE = "deploy.yaml"
_CURRENT_DIR = '.'


@click.command()
@click.option('--template-file', '-t', type=click.Path(exists=True), help="FAM template file or path about function config")
@click.option('--cos-bucket', type=str, help="COS bucket name")
@click.option('--output-template-file', '-o',
              type=click.Path(),
              help="FAM output template file or path",
              default=_DEFAULT_OUT_TEMPLATE_FILE,
              show_default=True)
def package(template_file, cos_bucket, output_template_file):
    '''
    Package a scf and upload to the cos
    '''
    package = Package(template_file, cos_bucket, output_template_file)
    package.do_package()


class Package(object):

    def __init__(self, template_file, cos_bucket, output_template_file):
        self.template_file = template_file
        self.cos_bucket = cos_bucket
        self.output_template_file = output_template_file
        self.check_params()
        self.resource = Resources(Context.get_template_data(self.template_file))

    def do_package(self):
        for ns_name, ns in vars(self.resource).items():
            for func_name, func in vars(ns).items():
                code_url = self._do_package_core(getattr(func, "CodeUri", ""))
                if "cos_bucket_name" in code_url:
                    setattr(func, "CosBucketName", code_url["cos_bucket_name"])
                    setattr(func, "CosObjectName", code_url["CosObjectName"])
                    click.secho("Upload function zip file '{}' to COS bucket '{}' success".
                                format(os.path.basename(code_url["cos_object_name"]),
                                       code_url["cos_bucket_name"]), fg="green")
                elif "zip_file" in code_url:
                    setattr(func, "LocalZipFile", code_url["zip_file"])

        yaml_dump(self.resource.to_json(), self.output_template_file)
        click.secho("Generate deploy file '{}' success".format(self.output_template_file), fg="green")

    def check_params(self):
        if not self.template_file:
            click.secho("FAM Template Not Found", fg="red")
            raise TemplateNotFoundException("Missing option --template-file".format(self.template_file))

    def _do_package_core(self, func_path):
        zipfile, zip_file_name = self._zip_func(func_path)
        code_url = dict()
        if self.cos_bucket:
            CosClient().upload_file2cos(bucket=self.cos_bucket, file=zipfile.read(),
                                        key=zip_file_name)
            code_url["cos_bucket_name"] = self.cos_bucket
            code_url["cos_object_name"] = "/" + zip_file_name
        else:
            code_url["zip_file"] = os.path.join(os.getcwd(), zip_file_name)

        return code_url

    def _zip_func(self, func_path):
        buff = BytesIO()
        if not os.path.exists(func_path):
            raise ContextException("Function file or path not found by CodeUri '{}'".format(func_path))

        zip_file_name = str(uuid.uuid1()) + '.zip'
        cwd = os.getcwd()
        os.chdir(func_path)
        with ZipFile(buff, mode='w', compression=ZIP_DEFLATED) as zip_object:
            for current_path, sub_folders, files_name in os.walk(_CURRENT_DIR):
                for file in files_name:
                    zip_object.write(os.path.join(current_path, file))

        os.chdir(cwd)
        buff.seek(0)
        buff.name = zip_file_name

        # a temporary support for upload func from local zipfile
        with open(zip_file_name, 'wb') as f:
            f.write(buff.read())
            buff.seek(0)
        # click.secho("Compress function '{}' to zipfile '{}' success".format(func_config.name, zip_file_name))

        return buff, zip_file_name

    def _get_code_abs_path(self, code_uri):

        if not self._cwd or self._cwd == _CURRENT_DIR:
            self._cwd = os.getcwd()

        self._cwd = os.path.abspath(self._cwd)

        code_path = code_uri
        if not os.path.isabs(code_uri):
            code_path = os.path.normpath(os.path.join(self._cwd, code_path))

        return code_path

    def _get_cwd(self):
        cwd = os.path.dirname(os.path.abspath(self._template_path))
        return cwd



from tcfcli.cmds.local.libs.docker.manager import ContainerManager
from tcfcli.cmds.local.libs.local.local_runtime import LocalRuntime
from tcfcli.common.user_exceptions import FunctionNotFound


class LocalRuntimeManager(object):
    def __init__(self,
                 function_provider,
                 cwd,
                 env_vars=None,
                 debug_context=None,
                 region=None,
                 docker_network_id=None,
                 skip_pull_image=False):

        self._provider = function_provider
        self._cwd = cwd
        self._env_vars = env_vars or {}
        self._region = region
        self._debug_context = debug_context

        self._container_manager = ContainerManager(docker_network_id, skip_pull_image)

    def invoke(self, func_name, event=None, stdout=None, stderr=None):
        local_runtime = LocalRuntime(func_config=self._get_func_config(func_name),
                                     env_vars=self._env_vars,
                                     cwd=self._cwd,
                                     debug_options=self.debug_options,
                                     container_manager=self._container_manager)

        local_runtime.invoke(event, stdout=stdout, stderr=stderr)

    def _get_func_config(self, func_name):
        func_config = self._provider.get(func_name)

        if not func_config:
            raise FunctionNotFound('function with name "%s" not found' % func_name)

        return func_config

    @property
    def debug_options(self):
        return self._debug_context



import copy
from tcfcli.libs.function.provider import TENCENT_FUNCTION_TYPE
from tcfcli.common.user_exceptions import InvalidDocumentException, InvalidTemplateException


class TemplateReimplemented(object):
    def __init__(self, template_dict):
        self._glob = None
        self._glob_settings = None
        self._resources = None
        self._template = copy.deepcopy(template_dict)

    def parse(self):
        try:
            self._validate()
            if self._glob_settings:
                self._merge_globals()

            return self._template
        except InvalidDocumentException as e:
            raise e

    def _validate(self):
        # basic validate, future will add restrict validate
        if not TemplateReimplemented._check_key_exist("Resources", self._template):
            raise InvalidTemplateException("'Resources' section is required")

        if len(self._template["Resources"]) < 1:
            raise InvalidTemplateException("Function not found in the template")

        self._resources = self._template["Resources"]
        for logic_id in self._resources:
            if "Type" not in self._resources[logic_id]:
                raise InvalidTemplateException("Resource '{}' define error: Lost 'Type' properties".format(logic_id))

            if self._resources[logic_id]["Type"] != TENCENT_FUNCTION_TYPE:
                raise InvalidTemplateException("Resource '{}' define error: Must define the resource 'Type' like as "
                                               "'{}'".format(logic_id, TENCENT_FUNCTION_TYPE))

            if not TemplateReimplemented._check_key_exist("Properties", self._resources[logic_id]):
                raise InvalidTemplateException("Resource '{}' define error: Must define the resource 'Properties'".
                                               format(logic_id))

            if TemplateReimplemented._check_key_exist("Events", self._resources[logic_id]["Properties"]):
                self._check_events_valid(logic_id, self._resources[logic_id]["Properties"]["Events"])

        if TemplateReimplemented._check_key_exist("Globals", self._template):
            self._glob = self._template["Globals"]
            if TemplateReimplemented._check_key_exist("Function", self._glob):
                self._glob_settings = self._glob["Function"]
                TemplateReimplemented._check_properties_valid(self._glob_settings, "Globals")

    def _merge_globals(self):
        for func_name in self._resources:
            func_settings = self._template["Resources"][func_name]["Properties"].copy()
            glob_settings = copy.deepcopy(self._glob_settings)
            glob_settings.update(func_settings)
            func_settings.update(glob_settings)
            self._template["Resources"][func_name]["Properties"] = func_settings

    @staticmethod
    def _check_key_exist(key, dict_map):
        if key not in dict_map or not isinstance(dict_map[key], dict) or\
                not dict_map[key]:
            return False
        return True

    @staticmethod
    def _check_properties_valid(dict_map, blob):
        if "Environment" in dict_map:
            TemplateReimplemented._check_env_valid(dict_map["Environment"], blob)

        if "VpcConfig" in dict_map:
            TemplateReimplemented._check_vpc_valid(dict_map["VpcConfig"], blob)

        if "FunctionName" in dict_map:
            raise InvalidTemplateException("Can not define 'FunctionName' in Globals block")

        if "CodeUri" in dict_map:
            raise InvalidTemplateException("Can not define 'CodeUri' in Globals block")

        if "Events" in dict_map:
            raise InvalidTemplateException("Can not define 'Events' in Globals block")

    @staticmethod
    def _check_env_valid(dict_map, blob):
        if not TemplateReimplemented._check_key_exist("Variables", dict_map):
            raise InvalidTemplateException("Has Environment but not define Variables in '{}'".format(blob))

        if len(dict_map["Variables"]) < 1:
            raise InvalidTemplateException("Environment values in '{}' not defined".format(blob))

    @staticmethod
    def _check_vpc_valid(dict_map, blob):
        if not TemplateReimplemented._check_key_exist("VpcId", dict_map):
            raise InvalidTemplateException("VpcId not found in '{}'".format(blob))

        if not TemplateReimplemented._check_key_exist("SubnetId", dict_map):
            raise InvalidTemplateException("SubnetTd not found in '{}'".format(blob))

    @staticmethod
    def _check_events_valid(func, dict_map):
        for event_id, config in dict_map.items():
            if "Type" not in config:
                raise InvalidTemplateException("Events 'Type' not found in Resource '{}'".format(func))

            if config["Type"] != 'Api':
                raise InvalidTemplateException("Events 'Type' not equal 'Api' in Resource '{}'".format(func))

            if "Properties" not in config:
                raise InvalidTemplateException("Events 'Properties' not found in Resource '{}'".format(func))

            if "Path" not in config["Properties"]:
                raise InvalidTemplateException("Api Path not found in Event from Resource '{}'".format(func))

            if "Method" not in config["Properties"]:
                raise InvalidTemplateException("Api Method not found in Event from Resource '{}'".format(func))

            if "StageName" in config["Properties"]:
                # TODO: if check?
                pass



import sys

PY_2 = 2
PY_3 = 3
pyv = sys.version_info
version = PY_2
if pyv[0] == PY_3:
    version = PY_3


def url_encoding(data):
    if version == PY_2:
        import urllib
        return urllib.quote(data)
    from urllib import parse
    return parse.quote(data)




import click
import platform
from tcfcli.common.user_config import UserConfig
version = platform.python_version()
if version >= '3':
    from functools import reduce

@click.command()
@click.option('--secret-id', is_flag=True, help="TencentCloudAPI  SecretId")
@click.option('--secret-key', is_flag=True, help="TencentCloudAPI  SecretKey")
@click.option('--region', is_flag=True,  help="TencentCloudAPI  Region")
@click.option('--appid', is_flag=True, help="TencentCloudAPI  Appid")
def get(**kwargs):
    uc = UserConfig()
    def set_true(k):
        kwargs[k] = True
    bools = [v for k, v in kwargs.items()]
    if not reduce(lambda x, y: bool(x or y), bools):
        list(map(set_true, kwargs))
    attrs = uc.get_attrs(kwargs)
    msg = "{} config:\n".format(UserConfig.API)
    for attr in sorted(attrs):
         msg += "{} = {}\n".format(attr, attrs[attr])
    click.secho(msg.strip())


@click.command()
@click.option('--secret-id', help="TencentCloudAPI  SecretId")
@click.option('--secret-key', help="TencentCloudAPI  SecretKey")
@click.option('--region', help="TencentCloudAPI  Region")
@click.option('--appid', help="TencentCloudAPI  Region")
def set(**kwargs):
    uc = UserConfig()
    uc.set_attrs(kwargs)
    uc.flush()


@click.group(name='configure')
def configure():
    """
    Configure your account parameters
    """
    pass

configure.add_command(get)
configure.add_command(set)



import os
import click
import platform

sysstr = platform.system()
if (sysstr == 'Linux' or 'Darwin'):
    home = os.path.expanduser('~')
    _USER_CONFIG_FILE = home + '/.tcli_config.ini'
else:
    _USER_CONFIG_FILE = ".tcli_config.ini"

version = platform.python_version()
if version >= '3':
    from configparser import ConfigParser
else:
    from ConfigParser import ConfigParser


class CliConfigParser(ConfigParser):

    def __init__(self, defaults=None):
        ConfigParser.__init__(self,defaults=None)
        #super(CliConfigParser, self).__init__(defaults)

    def optionxform(self, optionstr):
        return optionstr


class UserConfig(object):

    API = "API"

    def __init__(self):
        self.secret_id = 'None'
        self.secret_key = 'None'
        self.region = 'None'
        self.appid = 'None'
        self._load_config()

    def set_attrs(self, attrs):
        for k in attrs:
            if hasattr(self, self._name_attr2obj(k)) and attrs[k]:
                setattr(self, self._name_attr2obj(k), attrs[k])

    def get_attrs(self, attrs):
        objs = self._list_attrs()
        ret = {}
        for attr in [ k for k, v in attrs.items() if v]:
            obj = self._name_attr2obj(attr)
            if obj in objs:
                ret[self._name_obj2attr(attr)] = objs[obj]
        return ret

    def flush(self):
        self._dump_config()

    def _load_config(self):
        cf = CliConfigParser()
        if not cf.read(_USER_CONFIG_FILE):
            return
        if UserConfig.API not in cf.sections():
            return
        attrs = cf.options(UserConfig.API)

        for attr in attrs:
            if hasattr(self, self._name_attr2obj(attr)):
                setattr(self, self._name_attr2obj(attr), cf.get(UserConfig.API, attr))

    def _dump_config(self):
        objs = self._list_attrs()
        cf = CliConfigParser()

        cf.add_section(UserConfig.API)
        for obj in sorted(objs):
            cf.set(UserConfig.API, self._name_obj2attr(obj), objs[obj])

        with open(_USER_CONFIG_FILE, "w") as f:
            cf.write(f)

    def _list_attrs(self):
        attrs = vars(self)
        return {k : attrs[k]  for k in attrs if not k.startswith("_")}

    @staticmethod
    def _name_attr2obj(name):
        return name.replace("-", "_")

    @staticmethod
    def _name_obj2attr(name):
        return name.replace("_", "-")


if __name__ == '__main__':
    uc = UserConfig()
    #print uc.get_attrs({"secret-sssd": True, "region": False, "secret-key": "000000000"})
    #print uc.get_attrs({"secret-id": True, "region": True, "secret-key": "000000000"})
    #uc.set_attrs({"region":"asss-gz"})
    #uc.flush()
    print(uc.get_attrs({"region": True, "secret-key": "000000000"}))



import os
import yaml
from tcfcli.common.user_exceptions import ContextException
from tcfcli.libs.utils.yaml_parser import yaml_parse


class Template(object):

    @staticmethod
    def get_template_data(template_file):
        if not os.path.exists(template_file):
            return {}

        with open(template_file, 'r') as f:
            try:
                return yaml_parse(f.read())
            except (ValueError, yaml.YAMLError) as ex:
                raise ContextException("Parse template failed: {}".format(str(ex)))


from tcfcli.common.user_exceptions import InvalidOptionValue
from tcfcli.common import macro


class DebugContext(object):

    DEBUG_CMD = {
        macro.RUNTIME_NODEJS_610: macro.CMD_NODE,
        macro.RUNTIME_NODEJS_89: macro.CMD_NODE
    }

    def __init__(self, port, argv, runtime):
        self.debug_port = port
        self.debug_argv = argv
        self.runtime = runtime

    @property
    def is_debug(self):
        return self.debug_port is not None

    @property
    def cmd(self):
        if self.debug_port is None:
            return self.DEBUG_CMD[self.runtime]
        return None

    @property
    def argv(self):
        if self.debug_port is None:
            return []
        argv = []
        if self.debug_argv:
            argv.append(self.debug_argv)
        if self.runtime not in self.DEBUG_CMD.keys():
            raise InvalidOptionValue("Invalid runtime. [{}] support debug".format(",".join(self.DEBUG_CMD.keys())))

        if self.runtime == macro.RUNTIME_NODEJS_610:
            argv += self.debug_arg_node610
        elif self.runtime == macro.RUNTIME_NODEJS_89:
            argv += self.debug_arg_node89
        else:
            pass
        return argv

    @property
    def debug_arg_node610(self):
        return [
            "--inspect",
            "--debug-brk=" + str(self.debug_port),
            "--nolazy",
            "--max-old-space-size=2547",
            "--max-semi-space-size=150",
            "--expose-gc",
        ]

    @property
    def debug_arg_node89(self):
        return [
            "--inspect-brk=0.0.0.0:" + str(self.debug_port),
            "--nolazy",
            "--expose-gc",
            "--max-semi-space-size=150",
            "--max-old-space-size=2707",
        ]


import json

class ApigwEvent(object):
    def __init__(self,
                 method=None,
                 path=None,
                 body=None,
                 headers=None,
                 req_context=None,
                 queries_dict=None,
                 path_paras=None):

        if not isinstance(headers, dict) and headers is not None:
            raise TypeError('queries_dict is not dict or None' )

        if not isinstance(req_context, dict) and req_context is not None:
            raise TypeError('req_context is not dict or None' )

        if not isinstance(queries_dict, dict) and queries_dict is not None:
            raise TypeError('queries_dict is not dict or None' )

        if not isinstance(path_paras, dict) and path_paras is not None:
            raise TypeError('path_paras is not dict or None' )

        self._method = method
        self._path = path
        self._body = body
        self._headers = headers
        self._req_context = req_context
        self._queries_dict = queries_dict
        self._path_paras = path_paras

    def to_dict(self):
        self._event_dict = {
            'httpMethod': self._method,
            'path': self._path,
            'requestContext': self._req_context,
            'headersParameters': dict(self._headers) if self._headers else {},
            'pathParameters': dict(self._path_paras) if self._path_paras else {},
            'queryStringParameters': dict(self._queries_dict) if self._queries_dict else {},

        }

        if self._body:
            self._event_dict['body'] = self._body

        return self._event_dict

    def to_str(self):
        return json.dumps(self.to_dict())



import click
import os


_DEAFULT_TEMPLATE_FILE = 'template.[yaml|yml]'


def get_template_abspath(ctx, param, template_name):
    if template_name == _DEAFULT_TEMPLATE_FILE:
        template_name = 'template.yaml'

        tmp = 'template.yml'
        if os.path.exists(tmp):
            template_name = tmp

    return os.path.abspath(template_name)


def template_click_option():
    """
    Click Option for template option
    """
    return click.option('--template', '-t',
                        default=_DEAFULT_TEMPLATE_FILE,
                        type=click.Path(exists=True),
                        envvar="TCF_TEMPLATE_FILE",
                        callback=get_template_abspath,
                        show_default=True)


def invoke_common_options(f):
    invoke_options = [
        template_click_option(),

        click.option('--env-vars', '-n',
                     help='JSON file contains function environment variables.',
                     type=click.Path(exists=True)),

        click.option('--debug-port', '-d',
                     help='The port exposed for debugging. If specified, local container will start with debug mode.',
                     envvar="TCF_DEBUG_PORT"),

        click.option('--debugger-path',
                     help='The debugger path in host. If specified, the debugger will mounted into the function container.'),

        click.option('--debug-args',
                     help='Additional args to be passed the debugger.',
                     envvar="DEBUGGER_ARGS"),

        click.option('--docker-volume-basedir', '-v',
                     help='The basedir where TCF template locate in.',
                     envvar="TCF_DOCKER_VOLUME_BASEDIR"),

        click.option('--docker-network',
                     help='Specifies the name or id of an existing docker network which containers should connect to, '
                          'along with the default bridge network.',
                     envvar="TCF_DOCKER_NETWORK"),

        click.option('--log-file', '-l',
                     help='Path of logfile where send runtime logs to'),

        click.option('--skip-pull-image',
                     is_flag=True,
                     help='Specify whether CLI skip pulling or update docker images',
                     envvar="TCF_SKIP_PULL_IMAGE"),

        click.option('--region'),

    ]

    for option in reversed(invoke_options):
        option(f)

    return f


def service_common_options(port):
    def construct_options(f):
        service_options = [
            click.option('--host',
                         default="127.0.0.1",
                         help="Local hostname or IP address bind to (default: '127.0.0.1')"),
            click.option("--port", "-p",
                         default=port,
                         help="Local port number to listen on (default: '{}')".format(str(port)))
        ]

        for option in reversed(service_options):
            option(f)

        return f
    return construct_options


import click
from tcfcli.libs.tcsam.tcsam import Resources
from tcfcli.libs.function.context import Context


@click.command(name="validate")
@click.option('--template-file', '-t', type=click.Path(exists=True), help="TCF template file for deploy")
def validate(template_file):
    '''
    validate a scf template.
    '''
    resource = Resources(Context.get_template_data(template_file))


import yaml
import json


def yaml_dump(dict_to_dump, file_path):
    with open(file_path, "w") as f:
        yaml.safe_dump(dict_to_dump, f, default_flow_style=False)


def yaml_parse(yaml_file):
    try:
        return json.loads(yaml_file)
    except ValueError:
        return yaml.safe_load(yaml_file)



import io
import re
import os
from setuptools import setup, find_packages


def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', os.linesep)
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


def read_requirements(req='requirements.txt'):
    content = read(os.path.join(req))
    return [line for line in content.split(os.linesep)
            if not line.strip().startswith('#')]


def read_version():
    content = read(os.path.join(
        os.path.dirname(__file__), 'tcfcli', 'cmds', 'cli', '__init__.py'))
    return re.search(r"__version__ = '([^']+)'", content).group(1)


cmd_name = "tcf"

setup(
    name='tcf',
    version=read_version(),
    packages=find_packages(),
    description='This is a local tools for SCF.',
    long_description=open('README.md').read(),
    author='Tencent Cloud',
    url='https://github.com/tencentyun/tcfcli.git',
    maintainer_email="qcloud_middleware@qq.com",
    license="Apache License 2.0",
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*',
    entry_points={
        'console_scripts': [
            '{}=tcfcli.cmds.cli.cli:cli'.format(cmd_name)
        ]
    },
    install_requires=read_requirements('requirements.txt'),
    include_package_data=True,
)



from copy import deepcopy


class ApiProvider(object):

    _RESOURCE = 'Resources'
    _PROPERTIES = 'Properties'
    _EVENTS = 'Events'
    _PATH = 'Path'
    _METHOD = 'Method'

    _ANY_METHODS = ['GET', 'POST', 'PUT', 'DELETE', 'HEAD', 'ANY']

    def __init__(self, template):
        self._template = deepcopy(template)
        self._resources = self._template.get(self._RESOURCE, {})
        self.apis = self._extract_apis(self._resources)

    def get_all(self):
        for api in self.apis:
            yield api

    def _extract_apis(self, resources):
        apis = []

        for func_name, resource in resources.items():
            if self._EVENTS in resource[self._PROPERTIES]:
                for event, config in resource[self._PROPERTIES][self._EVENTS].items():
                    path = config[self._PROPERTIES].get(self._PATH, None)
                    method = config[self._PROPERTIES].get(self._METHOD, None)

                    api = API(path, method, func_name)
                    apis.append(api)

        return ApiProvider._normalize_apis(apis)

    @staticmethod
    def _normalize_apis(apis):
        res = []

        for api in apis:
            for normal_method in ApiProvider._normalize_method(api.method):
                res.append(API(path=api.path, method=normal_method, func_name=api.func_name))

        return res

    @staticmethod
    def _normalize_method(api_method):
        if api_method.upper() == 'ANY':
            for method in ApiProvider._ANY_METHODS:
                yield method
        else:
            yield api_method.upper()


class API(object):
    def __init__(self, path, method, func_name, ):
        self.path = path
        self.method = method
        self.func_name = func_name




from tcfcli.libs.tcsam.model import Model
from tcfcli.libs.tcsam.model import VARIABLE
from tcfcli.common.user_exceptions import InvalidTemplateException


class Environment(Model):

    def __init__(self, model):
        if not isinstance(model, dict) or VARIABLE not in model \
         or not isinstance(model[VARIABLE], dict):
            raise InvalidTemplateException("invalid")

        for m in model[VARIABLE]:
            if not isinstance(model[VARIABLE][m], str):
                raise InvalidTemplateException("%s invalid" % m)
            try:
                setattr(self, m, model[VARIABLE][m])
            except InvalidTemplateException as e:
                raise InvalidTemplateException("{proper} ".format(proper=m) + str(e))

    def to_json(self):
        return {VARIABLE: self._serialize()}



import click
from tcfcli.cmds.local.common.options import invoke_common_options, service_common_options
from tcfcli.cmds.local.common.invoke_context import InvokeContext
from tcfcli.cmds.local.libs.apigw.api_service import LocalApiService


@click.command(short_help='Set up a local service to simulate invoke by API event')
@service_common_options(3000)
@click.option("--static-dir", "-s",
              default="public",
              help="Any static assets (e.g. CSS/Javascript/HTML) files located in this directory "
                   "will be presented at /")
@invoke_common_options
def start_api(host, port, static_dir, template, env_vars, debug_port, debug_args, debugger_path,
              docker_volume_basedir, docker_network, log_file, skip_pull_image, region):

    do_start_api(host, port, static_dir, template, env_vars, debug_port, debug_args, debugger_path,
                 docker_volume_basedir, docker_network, log_file, skip_pull_image, region)


def do_start_api(host, port, static_dir, template, env_vars, debug_port, debug_args, debugger_path,
                 docker_volume_basedir, docker_network, log_file, skip_pull_image, region):

    with InvokeContext(template_file=template,
                       function_identifier=None,
                       env_vars_file=env_vars,
                       debug_port=debug_port,
                       debug_args=debug_args,
                       debugger_path=debugger_path,
                       docker_volume_basedir=docker_volume_basedir,
                       docker_network=docker_network,
                       log_file=log_file,
                       skip_pull_image=skip_pull_image,
                       region=region) as context:

        LocalApiService(invoke_context=context, port=port, host=host, static_dir=static_dir).start()



from tcfcli.common.user_exceptions import InvalidTemplateException
MUST = "must"
VALUE = "value"
DEFAULT = "default"
TYPE = "Type"
PROPERTY = "Properties"
VARIABLE = "Variables"
RESOURCE = "Resources"
FUNCTION = "Function"
GLOBAL = "Globals"
CODE_URI = "CodeUri"
ENVIRONMENT = "Environment"
HANDLER = "Handler"
MEMORY_SIZE = "MemorySize"
RUNTIME = "Runtime"
TIMEOUT = "Timeout"
mint = lambda x: int(x)
mstr = lambda x: str(x)
mbool = lambda x: bool(x)
mmap = lambda x: x


class Model(object):

    TYPE_PREFIX = ""

    def _serialize(self):
        """Get all params which are not None if None is not allowed."""
        js = {}
        propers = vars(self)

        for k, v in propers.items():
            if isinstance(v, Model):
                js[k] = v.to_json()
            else:
                js[k] = v
        return js

    def to_json(self):
        return self._serialize()


class ProperModel(Model):
    PROPER = {
    }

    def __init__(self, model):
        if not isinstance(model, dict) or PROPERTY not in model:
            raise InvalidTemplateException("%s invalid" % PROPERTY)
        if not isinstance(model[PROPERTY], dict):
            if model[PROPERTY] is None:
                model[PROPERTY] = {}
            else:
                raise InvalidTemplateException("%s invalid" % PROPERTY)

        for m in model[PROPERTY]:
            if m not in self.PROPER:
                raise InvalidTemplateException("%s invalid" % m)
            try:
                setattr(self, m, self.PROPER[m][TYPE](model[PROPERTY][m]))
            except InvalidTemplateException as e:
                raise InvalidTemplateException("{proper} ".format(proper=m) + str(e))

        proper = [x for x in self.PROPER if self.PROPER[x][MUST]]
        for p in proper:
            v = getattr(self, p, None)
            if v is None:
                raise InvalidTemplateException("%s invalid" % p)
            elif isinstance(self.PROPER[p][VALUE], tuple) and v not in self.PROPER[p][VALUE]:
                raise InvalidTemplateException("%s invalid" % p)

    def to_json(self):
        return {TYPE: self.TYPE_PREFIX + self.__class__.__name__, PROPERTY: self._serialize()}


class TypeModel(Model):
    AVA_TYPE = {
    }

    def __init__(self, model, dft={}):
        if not isinstance(model, dict):
            raise InvalidTemplateException("invalid")
        for m in model:
            mod = model[m]
            if not isinstance(mod, dict) or TYPE not in mod:
                raise InvalidTemplateException("%s invalid" % m)
            type = mod[TYPE]
            del mod[TYPE]
            if type not in self.AVA_TYPE:
                raise InvalidTemplateException("%s invalid" % m)
            try:
                setattr(self, m, self.AVA_TYPE[type](mod, dft))
            except InvalidTemplateException as e:
                raise InvalidTemplateException("{proper} ".format(proper=m) + str(e))

    def to_json(self):
        t = self._serialize()
        t[TYPE] = self.TYPE_PREFIX + self.__class__.__name__
        return t


import os
import sys
import json
import shutil
import tempfile
import zipfile
import click
import signal
import threading
from contextlib import contextmanager

from tcfcli.cmds.local.libs.docker.container import Container
from tcfcli.common.user_exceptions import InvalidEnvParameters


class LocalRuntime(object):
    _BASE_IMAGE_NAME = 'hub.tencentyun.com/qcloud_scf/scf'
    _GOLANG_SERVER_PORT = 33300
    _WORK_DIR = '/var/user'
    _CURRENT_DIR = '.'
    _BLANK = ''

    _ARCHIVE_FORMATS = ('.zip', '.ZIP')

    _RUNTIME_LIST = [
        'python2.7',
        'python3.6',
        'go1',
        'php5',
        'php7',
        'nodejs6.10',
        'nodejs8.9',
        'java8',
    ]

    def __init__(self, func_config, env_vars=None, cwd=None, debug_options=None, container_manager=None):
        self._func_config = func_config
        self._env_vars = env_vars
        self._cwd = cwd
        self._debug_options = debug_options
        self._container_manager = container_manager

        if self.get_runtime() not in self._RUNTIME_LIST:
            print(self.get_runtime())
            raise ValueError('Not support runtime {}'.format(self.get_runtime()))

        self._container = None

    def invoke(self, event=None, stdout=None, stderr=None):
        image = self.get_image()
        cmd = [self.get_handler()]
        code_abs_path = self.get_code_abs_path()
        memory = self.get_memory()
        entry = self.get_entry_point()
        ports ={self._debug_options.debug_port: self._debug_options.debug_port} \
            if self._debug_options else None
        envs = self.get_envs(event)
        timer = None

        with self._get_code(code_abs_path) as code_dir:
            self._container = Container(image=image,
                                        cmd=cmd,
                                        work_dir=self._WORK_DIR,
                                        host_dir=code_dir,
                                        mem=memory,
                                        env_vars=envs,
                                        entrypoint=entry,
                                        ports=ports)

            try:
                self._container_manager.run(self._container)

                timer = self._wait_timeout(self._container, self.get_timeout(), bool(self._debug_options))

                self._container.get_logs(stdout=stdout, stderr=stderr)

            except KeyboardInterrupt:
                click.secho('Abort function execution')
            except Exception as err:
                click.secho('Invoke error:%s' % str(err))

            finally:
                if timer:
                    timer.cancel()
                self._container.delete()

    def get_func_name(self):
        return self._func_config.name

    def get_image(self):
        runtime = self.get_runtime()
        return '%s:%s' % (self._BASE_IMAGE_NAME, runtime)

    def get_runtime(self):
        return self._func_config.runtime.lower()

    def get_handler(self):
        res = self._func_config.handler.split('.', 1)
        if len(res) > 1:
            return res[0] + ':' + res[1]
        return res[0]

    def get_golang_port(self):
        return self._GOLANG_SERVER_PORT

    def get_code_uri(self):
        return self._func_config.codeuri

    def get_memory(self):
        return self._func_config.memory

    def get_timeout(self):
        return int(self._func_config.timeout)

    def get_entry_point(self):
        if not self._debug_options:
            return None

        runtime = self.get_runtime()
        if runtime not in ["nodejs6.10", "nodejs8.9"]:
            raise InvalidEnvParameters("Debugging is not currently supported for {}".format(runtime))

        debug_port = self._debug_options.debug_port
        debug_args = self._debug_options.debug_args
        debug_args_list = []
        if debug_args:
            debug_args_list = debug_args.split(" ")

        entrypoint = None
        if runtime == "nodejs6.10":
            entrypoint = ["/var/lang/node6/bin/node"] \
                   + debug_args_list \
                   + [
                       "--inspect",
                       "--debug-brk=" + str(debug_port),
                       "--nolazy",
                       "--max-old-space-size=2547",
                       "--max-semi-space-size=150",
                       "--max-executable-size=160",
                       "--expose-gc",
                       "/var/runtime/node6/bootstrap.js",
                   ]
        elif runtime == "nodejs8.9":
            entrypoint = ["/var/lang/node8/bin/node"] \
                    + debug_args_list \
                    + [
                        "--inspect-brk=0.0.0.0:" + str(debug_port),
                        "--nolazy",
                        "--expose-gc",
                        "--max-semi-space-size=150",
                        "--max-old-space-size=2707",
                        "/var/runtime/node8/bootstrap.node8.js",
                    ]
        return entrypoint

    def get_code_abs_path(self):
        # return the code path based on current working directory

        if not self._cwd or self._cwd == self._CURRENT_DIR:
            self._cwd = os.getcwd()

        self._cwd = os.path.abspath(self._cwd)

        code_path = self.get_code_uri()
        if not os.path.isabs(self.get_code_uri()):
            code_path = os.path.normpath(os.path.join(self._cwd, code_path))

        return code_path

    def get_envs(self, event=None):
        func_name = self._func_config.name
        variables = None
        if self._func_config.environment and isinstance(self._func_config.environment, dict) and 'Variables' in self._func_config.environment:
            variables = self._func_config.environment['Variables']

        for env_var in self._env_vars.values():
            if not isinstance(env_var, dict):
                raise InvalidEnvParameters('Environment variables format is invalid, format must like as {FunctionName: {key: value} JSON pairs}')

        overrides = self._env_vars.get(func_name, None)

        return self.generate_runtime_envs(variables, overrides, event)

    def generate_runtime_envs(self, template_env, overrides, event):
        template_env_values = template_env or {}
        overrides_values = overrides or {}

        env_set = {
            'SCF_LOCAL': 'true',
            '_SCF_SERVER_PORT': str(self.get_golang_port()),
            '_LAMBDA_SERVER_PORT': str(self.get_golang_port()),
            'SCF_FUNCTION_NAME': self.get_func_name(),
            'SCF_FUNCTION_MEMORY_SIZE': str(self.get_memory()),
            'SCF_FUNCTION_TIMEOUT': str(self.get_timeout()),
            'SCF_FUNCTION_HANDLER': self.get_handler(),

            'SCF_EVENT_BODY': '{}'
        }

        environ = {}
        for key, value in template_env_values.items():
            if key in overrides_values:
                value = overrides_values[key]

            environ[key] = value
            env_set[key] = self._stringfy(value)

        env_set['SCF_FUNCTION_ENVIRON'] = json.dumps(environ)

        if event:
            env_set['SCF_EVENT_BODY'] = event

        return env_set

    def _stringfy(self, value):
        if isinstance(value, (dict, tuple, list)) or value is None:
            res = self._BLANK
        elif value is True:
            res = 'true'
        elif value is False:
            res = 'false'
        elif sys.version_info.major > 2:
            res = str(value)
        elif not isinstance(value, unicode):
            res = str(value)
        else:
            res = value

        return res

    @contextmanager
    def _get_code(self, code_abs_path):
        tmp_code_path = None

        try:
            if os.path.isfile(code_abs_path) and code_abs_path.endswith(self._ARCHIVE_FORMATS):
                tmp_code_path = self._get_tmp_code_path(code_abs_path)
                yield tmp_code_path
            else:
                yield code_abs_path
        finally:
            if tmp_code_path:
                shutil.rmtree(tmp_code_path)

    def _get_tmp_code_path(self, code_abs_path):
        tmp_dir = tempfile.mkdtemp()

        if os.name == 'posix':
            os.chmod(tmp_dir, 0o755)

        self._unzip(code_abs_path, tmp_dir)

        return os.path.realpath(tmp_dir)

    def _unzip(self, source_file, target_dir):
        with zipfile.ZipFile(source_file, 'r') as f:
            for file in f.namelist():
                f.extract(file, target_dir)

    def _wait_timeout(self, container=None, timeout=None, isdebug=False):
        if not timeout:
            timeout = 3

        def signal_handler(sig, frame):
            container.delete()

        def stop_container():
            click.secho('Function "%s" timeout after %d seconds' % (self.get_func_name(), timeout))
            container.delete()

        if isdebug:
            signal.signal(signal.SIGTERM, signal_handler)
            return

        timer = threading.Timer(timeout, stop_container)
        timer.start()

        return timer


import io
import os
import sys
import json
import uuid
import base64
import logging

from flask import Flask, Response, request
from tcfcli.cmds.local.libs.apigw.path_converter import PathConverter
from tcfcli.cmds.local.libs.events.api import ApigwEvent
from tcfcli.cmds.local.libs.apigw.error_response import ErrorResponse

logger = logging.getLogger(__name__)


class LocalService(object):
    def __init__(self, routes_list, runtime_manager, static_dir=None, port=None, host=None, stderr=None):
        self._routes_list = routes_list
        self._runtime_manager = runtime_manager
        self._static_dir = static_dir
        self._port = port
        self._host = host
        self._stderr = stderr

        self._route_map = {}
        self._server = None

    def create(self):
        self._server = Flask(__name__, static_url_path='', static_folder=self._static_dir)

        for route in self._routes_list:
            path = PathConverter.convert_path_to_flask(route.path)
            self._save_route_map(path, route)
            self._server.add_url_rule(path,
                                      endpoint=path,
                                      methods=route.method,
                                      view_func=self._request_handler,
                                      provide_automatic_options=False)

    def listen(self):
        self.create()

        if not self._server:
            raise RuntimeError('Local service has not been created before listening')

        os.environ['WERKZEUG_RUN_MAIN'] = 'true'

        self._server.run(threaded=True, host=self._host, port=self._port)

    def _request_handler(self, **kwargs):
        key = request.path + ':' + request.method

        if key not in self._route_map.keys():
            raise RuntimeError('Not found the function bind to the route')

        try:
            event = self._generate_api_event(request)
        except UnicodeDecodeError:
            return ErrorResponse.InternalError()

        stdout = io.BytesIO()
        try:
            self._runtime_manager.invoke(self._route_map[key].func_name,
                                         event,
                                         stdout=stdout,
                                         stderr=self._stderr)
        except:
            return ErrorResponse.InternalError()

        return_value, log_list= self._get_ouput(stdout)

        if self._stderr and log_list:
            for line in log_list:
                if sys.version_info.major > 2:
                    self._stderr.write(bytes(line + '\n', encoding='utf-8'))
                else:
                    self._stderr.write(bytes(line + '\n').encode("utf-8"))

        try:
            status_code, headers, body = self._parse_output(return_value)
        except TypeError as ex:
            logging.error('parse return value error: {}'.format(str(ex)))
            return ErrorResponse.InvalidResponseFormat()

        return self._response(status_code, headers, body)

    def _save_route_map(self, path=None, route=None):
        if not route or not path:
            return

        for method in route.method:
            key = '{}:{}'.format(path, method)
            self._route_map[key] = route

    @staticmethod
    def _generate_api_event(request):
        req_context = {
            'path': PathConverter.convert_path_to_api_gateway(request.endpoint),
            'httpMethod': request.method,
            'requestId': str(uuid.uuid1()),
            'sourceIp': request.remote_addr,
            'stage': 'prod',
        }

        headers = dict(request.headers)

        body = request.get_data()
        if body:
            body = body.decode('utf-8')

        queries = LocalService._generate_query_dict(request)

        event = ApigwEvent(method=request.method,
                           path=request.path,
                           path_paras=request.view_args,
                           body=body,
                           headers=headers,
                           req_context=req_context,
                           queries_dict=queries)

        return event.to_str()

    @staticmethod
    def _generate_query_dict(request):
        query_dict = {}

        for query_key, query_values in request.args.lists():
            query_values_num = len(query_values)

            if not query_values_num:
                query_dict[query_key] = ""
            else:
                query_dict[query_key] = query_values[-1]

        return query_dict

    @staticmethod
    def _get_ouput(stdout_stream):
        output = stdout_stream.getvalue().decode('utf-8').strip('\n').split('\n')

        output_length = len(output)
        log_list = None
        if output_length > 1:
            log_list = output[:output_length - 1]

        return_value = output[-1]

        return return_value, log_list

    @staticmethod
    def _parse_output(return_vlue):
        try:
            return_json = json.loads(return_vlue)
        except:
            raise TypeError('return value must be a valid json')

        if not isinstance(return_json, dict):
            raise TypeError('return value must be a valid json, not "{}"'.format(type(return_json)))

        status_code = return_json.get('statusCode') or None
        headers = return_json.get('headers') or None
        body = return_json.get('body') or ''
        is_base64_encoded = return_json.get('isBase64Encoded') or False

        if not isinstance(status_code, int) or status_code <= 0:
            raise TypeError('statusCode must be positive int')

        if not isinstance(headers, dict) or headers is None:
            raise TypeError('headers format is wrong')

        if is_base64_encoded:
            body = base64.b64decode(body)

        if 'Content-type' not in headers:
            headers['type'] = 'application/json'

        return status_code, headers, body

    @staticmethod
    def _response(status_code, headers, body):
        response = Response(body)
        response.headers = headers
        response.status_code = status_code

        return response


class Route(object):
    def __init__(self, method, path, func_name):
        self.method = method
        self.path = path
        self.func_name = func_name



import json
import sys
import docker
import os

from tcfcli.libs.function.context import Context as FuncContext
from tcfcli.libs.function.fam_function_provider import ScfFunctionProvider
from tcfcli.libs.tcsam.tcsam import Resources
from tcfcli.libs.tcsam import model
from tcfcli.common.user_exceptions import InvokeContextException
from tcfcli.cmds.local.libs.local.local_runtime_manager import LocalRuntimeManager
from tcfcli.cmds.local.libs.local.debug_context import DebugContext


class InvokeContext(object):
    def __init__(self,
                 template_file,
                 function_identifier=None,
                 env_vars_file=None,
                 debug_port=None,
                 debug_args=None,
                 debugger_path=None,
                 docker_volume_basedir=None,
                 docker_network=None,
                 log_file=None,
                 skip_pull_image=None,
                 region=None,
                 namespace=None):

        self._template_file = template_file
        self._function_identifier = function_identifier
        self._env_vars_file = env_vars_file
        self._debug_port = debug_port
        self._debug_args = debug_args
        self._debugger_path = debugger_path
        self._docker_volume_basedir = docker_volume_basedir
        self._docker_network = docker_network
        self._log_file = log_file
        self._skip_pull_image = skip_pull_image
        self._region = region
        self.namespace = namespace

        self._template_dict = None
        self._function_provider = None
        self._env_vars = None
        self._log_file_fp = None
        self._debug_context = None

    def __enter__(self):
        template_dict = Resources(FuncContext().get_template_data(self._template_file)).to_json()
        resource = template_dict.get(model.RESOURCE, {})
        ns = resource.get(self.namespace, {})
        if not ns:
            raise InvokeContextException("You must provide a namespace identifier,default is 'default'")
        if model.TYPE in ns:
            del ns[model.TYPE]
        self._template_dict = {model.RESOURCE: ns}
        self._function_provider = ScfFunctionProvider(template_dict=self._template_dict)
        self._env_vars = self._get_env_vars(self._env_vars_file)
        self._log_file_fp = self._get_log_file(self._log_file)
        self._debug_context = self._get_debug_context(self._debug_port, self._debug_args, self._debugger_path)

        self._check_docker()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._log_file_fp:
            self._log_file_fp.close()

    def get_cwd(self):
        if self._docker_volume_basedir:
            cwd = self._docker_volume_basedir
        else:
            cwd = os.path.dirname(os.path.abspath(self._template_file))

        return cwd

    @property
    def local_runtime_manager(self):
        return LocalRuntimeManager(function_provider=self._function_provider,
                                   cwd=self.get_cwd(),
                                   env_vars=self._env_vars,
                                   debug_context=self._debug_context,
                                   region=self._region,
                                   docker_network_id=self._docker_network,
                                   skip_pull_image=self._skip_pull_image)

    @property
    def template(self):
        return self._template_dict

    @property
    def functions_name(self):
        if self._function_identifier:
            return self._function_identifier

        functions = [f for f in self._function_provider.get_functions()]
        if len(functions) == 1:
            return functions[0].name

        function_names = [f.name for f in functions]

        raise InvokeContextException("You must provide a function identifier (function's Logical ID in the template). "
                                     "Possible options in your template: {}".format(function_names))

    @property
    def stdout(self):
        if self._log_file_fp:
            return self._log_file_fp

        byte_stdout = sys.stdout

        if sys.version_info.major > 2:
            byte_stdout = sys.stdout.buffer

        return byte_stdout

    @property
    def stderr(self):
        if self._log_file_fp:
            return self._log_file_fp

        byte_stderr = sys.stderr

        if sys.version_info.major > 2:
            byte_stderr = sys.stderr.buffer

        return byte_stderr

    @staticmethod
    def _get_env_vars(env_vars_file):
        if env_vars_file is None:
            return None

        try:
            with open(env_vars_file, 'r') as fp:
                return json.load(fp)
        except Exception as e:
            raise InvokeContextException('read environment from file {} failed: {}'.format(env_vars_file, str(e)))

    @staticmethod
    def _get_log_file(log_file):
        if log_file is None:
            return

        return open(log_file, 'wb')

    @staticmethod
    def _get_debug_context(debug_port, debug_paras, debugger_path):
        return DebugContext(debug_port, debugger_path, debug_paras)

    @staticmethod
    def _check_docker(docker_client=None):
        docker_client = docker_client or docker.from_env()

        try:
            docker_client.ping()
        except Exception as e:
            raise InvokeContextException('Docker not found, please install it. %s' % e)



import click
from tcfcli.cmds.local.generate_event.events_metadata import EVENTS_METADATA
from tcfcli.cmds.local.generate_event.generate_event_action import GenerateEventAction


class GenerateEventService(click.MultiCommand):
    '''
    this is a test
    '''

    def get_command(self, ctx, cmd):
        if cmd not in EVENTS_METADATA:
            return None
        return GenerateEventAction(cmd, EVENTS_METADATA[cmd])

    def list_commands(self, ctx):
        return sorted(EVENTS_METADATA.keys())



import docker
from .utils import to_posix_path
from .attach_api import attach


class Container(object):

    _STDOUT_FRAME_TYPE = 1
    _STDERR_FRAME_TYPE = 2

    def __init__(self, image,
                 cmd,
                 work_dir,
                 host_dir,
                 mem=None,
                 env_vars=None,
                 entrypoint=None,
                 ports=None,
                 docker_client=None,
                 network_id=None,
                 container_opts=None,
                 additional_volumes=None):

        self._image = image
        self._cmd = cmd
        self._memory = mem
        self._env_vars = env_vars
        self._working_dir = work_dir
        self._host_dir = host_dir
        self._ports = ports
        self._entrypoint = entrypoint
        self._network_id = network_id

        self._docker_client = docker_client or docker.from_env()
        self.id = None

    def create(self):
        if self.is_exist():
            raise Exception('container has been created, can not create again.')

        kwargs = {
            "command": self._cmd,
            "working_dir": self._working_dir,
            "volumes": {
                self._host_dir: {
                    "bind": self._working_dir,
                    "mode": "ro"
                }
            },
            "tty": False
        }

        kwargs["volumes"] = {to_posix_path(host_dir): mount for host_dir, mount in kwargs["volumes"].items()}

        if self._env_vars:
            kwargs["environment"] = self._env_vars

        if self._ports:
            kwargs["ports"] = self._ports

        if self._entrypoint:
            kwargs["entrypoint"] = self._entrypoint

        if self._memory:
            kwargs["mem_limit"] = "{}m".format(self._memory)

        container_instance = self._docker_client.containers.create(self._image, **kwargs)
        self.id = container_instance.id

        if self._network_id:
            network = self._docker_client.networks.get(self._network_id)
            network.connect(self.id)

        return self.id

    def start(self):
        container_instance = self._docker_client.containers.get(self.id)

        container_instance.start()

    def delete(self):
        if not self.is_exist():
            return

        try:
            self._docker_client.containers.get(self.id).remove(force=True)
        except docker.errors.NotFound:
            pass
        except docker.errors.APIError as e:
            err_msg = str(e)
            if err_msg.find('is already in progress') < 0 or err_msg.find('removal of container') < 0:
                raise e
        except docker.errors.DockerException as e:
            raise Exception('delete container %s failed, error: %s' % (self.id, str(e)))

        self.id = None

    def is_exist(self):
        return self.id is not None

    def get_logs(self, stdout=None, stderr=None):
        if not stdout and not stderr:
            return

        if not self.is_exist():
            raise Exception('can not get logs, container does not exist')

        container_instance = self._docker_client.containers.get(self.id)

        logs_itr = attach(self._docker_client,
                          container=container_instance,
                          stdout=True,
                          stderr=True,
                          logs=True)

        self._write_container_output(logs_itr, stdout=stdout, stderr=stderr)

    @staticmethod
    def _write_container_output(output_itr, stdout=None, stderr=None):
        for frame_type, data in output_itr:
            if frame_type == Container._STDOUT_FRAME_TYPE and stdout:
                stdout.write(data)
            elif frame_type == Container._STDERR_FRAME_TYPE and stderr:
                stderr.write(data)
            else:
                pass

    @property
    def image(self):
        return self._image
